@@ -1,9 +1,11 @@
 # -*- coding: utf-8 -*-
 
-from aristaflow.configuration import Configuration
-from aristaflow.client_service import AristaFlowClientService
+# AristaFlow REST Libraries
 from aristaflow.client_platform import AristaFlowClientPlatform
-from aristaflow.worklist_service import WorklistService
-from aristaflow.process_service import ProcessService
+from aristaflow.client_service import AristaFlowClientService
+from aristaflow.configuration import Configuration
 from aristaflow.delegation_service import DelegationService
+from aristaflow.process_service import ProcessService
 from aristaflow.remote_iterator_handler import RemoteIteratorHandler
+from aristaflow.service_provider import ServiceProvider
+from aristaflow.worklist_service import WorklistService
