@@ -1,6 +1,6 @@
 # Default Python Libraries
 import base64
-from typing import List, Type, TypeVar
+from typing import Dict, List, Type, TypeVar
 
 # AristaFlow REST Libraries
 import af_execution_manager
@@ -14,26 +14,25 @@ from af_org_model_manager.models.auth_data_user_name import AuthDataUserName
 from af_org_model_manager.models.authentication_data import AuthenticationData
 from af_org_model_manager.models.client_session_details import ClientSessionDetails
 from af_org_model_manager.models.qualified_agent import QualifiedAgent
-from af_remote_html_runtime_manager.api.runtime_manager_api import RuntimeManagerApi
 from af_remote_html_runtime_manager.api.synchronous_activity_starting_api import (
     SynchronousActivityStartingApi,
 )
-from af_remote_html_runtime_manager.models.activity_rest_callback_data import (
-    ActivityRestCallbackData,
-)
-from af_remote_html_runtime_manager.models.ebp_instance_reference import EbpInstanceReference
-from af_remote_html_runtime_manager.models.gui_context import GuiContext
 from af_worklist_manager.models.af_activity_reference import AfActivityReference
 from af_worklist_manager.models.worklist_item import WorklistItem
+from af_runtime_service import SimpleSessionContext
 from aristaflow.absence_service import AbsenceService
 from aristaflow.delegation_service import DelegationService
 from aristaflow.execution_history_service import ExecutionHistoryService
-from aristaflow.html_gui_context import HtmlGuiContext
+from aristaflow.image_renderer_service import ImageRendererService
 from aristaflow.org_model_service import OrgModelService
 from aristaflow.process_service import ProcessService
 from aristaflow.service_provider import ServiceProvider
+from aristaflow.activity_context import ActivityContext
+from .abstract_service import AbstractService
 
+from .activity_service import ActivityService
 from .configuration import Configuration
+from .remote_html_service import RemoteHtmlService, HtmlSignalHandler
 from .worklist_service import WorklistService
 
 
@@ -56,7 +55,11 @@ class AristaFlowClientService(object):
     __delegation_service: DelegationService = None
     __absence_service: AbsenceService = None
     __execution_history_service: ExecutionHistoryService = None
+    __image_renderer_service: ImageRendererService = None
     __org_model_service: OrgModelService = None
+    __activity_service: ActivityService = None
+    __remote_html_service: RemoteHtmlService = None
+    __all_services: List[AbstractService] = []
 
     def __init__(
         self, configuration: Configuration, user_session: str, service_provider: ServiceProvider
@@ -166,105 +169,84 @@ class AristaFlowClientService(object):
         self.__csd = csd
 
     def is_html_activity(self, item: WorklistItem) -> bool:
-        return (
-            item.act_ref.gui_context_id == "HTMLContext"
-            or item.act_ref.executable_component_name == "de.aristaflow.form.Form"
-            or item.act_ref.executable_component_name == "de.aristaflow.form.GeneratedForm"
-        )
+        return RemoteHtmlService.is_html_activity(item)
 
     @property
     def worklist_service(self):
         if self.__worklist_service is None:
             self.__worklist_service = WorklistService(self.__service_provider, self.__af_conf)
+            self.__all_services.append(self.__worklist_service)
         return self.__worklist_service
 
     @property
     def process_service(self):
         if self.__process_service is None:
             self.__process_service = ProcessService(self.__service_provider, self.__af_conf)
+            self.__all_services.append(self.__process_service)
         return self.__process_service
 
     @property
     def delegation_service(self):
         if self.__delegation_service is None:
             self.__delegation_service = DelegationService(self.__service_provider)
+            self.__all_services.append(self.__delegation_service)
         return self.__delegation_service
 
     @property
     def absence_service(self):
         if self.__absence_service is None:
             self.__absence_service = AbsenceService(self.__service_provider)
+            self.__all_services.append(self.__absence_service)
         return self.__absence_service
 
     @property
     def execution_history_service(self):
         if self.__execution_history_service is None:
             self.__execution_history_service = ExecutionHistoryService(self.__service_provider)
+            self.__all_services.append(self.__execution_history_service)
         return self.__execution_history_service
 
+    @property
+    def image_renderer_service(self):
+        if self.__image_renderer_service is None:
+            self.__image_renderer_service = ImageRendererService(self.__service_provider)
+            self.__all_services.append(self.__image_renderer_service)
+        return self.__image_renderer_service
+
     @property
     def org_model_service(self):
         if self.__org_model_service is None:
             self.__org_model_service = OrgModelService(self.__service_provider)
+            self.__all_services.append(self.__org_model_service)
         return self.__org_model_service
 
-    def start_html_activity(self, item: WorklistItem, callback_url: str):
+    @property
+    def activity_service(self):
+        if self.__activity_service is None:
+            self.__activity_service = ActivityService(self.__service_provider)
+            self.__all_services.append(self.__activity_service)
+        return self.__activity_service
+
+    @property
+    def remote_html_service(self):
+        if self.__remote_html_service is None:
+            self.__remote_html_service = RemoteHtmlService(self.__service_provider)
+            self.__all_services.append(self.__remote_html_service)
+        return self.__remote_html_service
+
+    def start_html_activity(self, item: WorklistItem, callback_url: str = None,
+                            signal_handler: HtmlSignalHandler = None):
         """
         Starts the given HTML GUI worklist item using the Remote HTML Runtime Manager
         """
-        if item is None:
-            raise Exception("No worklist item provided")
-        # accept user form and HTMLContext based activities
-        if not (self.is_html_activity(item)):
-            raise Exception(f"Not an HTML activity: {item.act_ref.gui_context_id}")
-        if item.state == "STARTED":
-            raise Exception("Item is already started")
-        sas = self.get_html_activity_starting()
-        gc: GuiContext
-        ar: AfActivityReference = item.act_ref
-        # print('Starting activity...')
-        ebp_ir: EbpInstanceReference = EbpInstanceReference(
-            ar.type,
-            ar.instance_id,
-            ar.instance_log_id,
-            ar.base_template_id,
-            ar.node_id,
-            ar.node_iteration,
-            ar.execution_manager_uris,
-            ar.runtime_manager_uris,
-        )
-        # "AVAILABLE", "ASSIGNED", "STARTED", "SUSPENDED", "ENQUIRED"
-        if item.state == "AVAILABLE" or item.state == "ASSIGNED":
-            if callback_url is not None:
-                cb_data = ActivityRestCallbackData(
-                    sub_class="ActivityRestCallbackData",
-                    notification_callback=callback_url,
-                    activity=ebp_ir,
-                )
-                gc = sas.start_activity_callback(body=cb_data)
-            else:
-                gc = sas.start_activity(body=ebp_ir)
-        else:
-            if callback_url is not None:
-                cb_data = ActivityRestCallbackData(
-                    sub_class="ActivityRestCallbackData",
-                    notification_callback=callback_url,
-                    activity=ebp_ir,
-                )
-                gc = sas.resume_activity_callback(body=cb_data)
-            else:
-                gc = sas.resume_activity(body=ebp_ir)
-        return HtmlGuiContext(gc)
+        return self.remote_html_service.start_html_activity(item, callback_url=callback_url,
+                                                            signal_handler=signal_handler)
 
     def get_html_activity_starting(self) -> SynchronousActivityStartingApi:
         """
         Returns the Remote HTML Runtime Manager Synchronous Activity Starting, ensuring logon to the Runtime Manager
         """
-        sas: SynchronousActivityStartingApi = self.get_service(SynchronousActivityStartingApi)
-        rm: RuntimeManagerApi = self.get_service(RuntimeManagerApi)
-        # always logon again, since the server might have been restarted in the meantime
-        rm.logon(body=self.__csd)
-        return sas
+        return self.remote_html_service.get_html_activity_starting()
 
     def reset_activity(self, item: WorklistItem):
         """Resets the given worklist item."""
@@ -290,11 +272,38 @@ class AristaFlowClientService(object):
         """Deserialize data using the given class of the generated OpenAPI models."""
         return self.__service_provider.deserialize(data, klass)
 
-    def serialize(self, obj) -> str:
+    def serialize(self, obj) -> Dict:
         """Serialize REST model object"""
         return self.__service_provider.serialize(obj)
 
+    def parse_ac_to_dict(self, ac: ActivityContext) -> dict:
+        """
+        Parses the context data required to recreate an ActivityContext
+        to a json serializable dict
+        """
+        return {
+            "token": ac.token,
+            "ssc": self.serialize(ac.ssc)
+        }
+
+    def get_ac_from_dict(self,context_data: dict) -> ActivityContext:
+        """
+        Creates a new ActivityContext from a dict provided by parse_ac_to_dict
+        """
+        ac = ActivityContext()
+        ac.token = context_data["token"]
+        ac.ssc = self.deserialize(context_data["ssc"], SimpleSessionContext)
+        return ac
+
     @property
     def autostart_timeout_seconds(self) -> int:
         """Wait time in seconds for auto start signals"""
         return self.__af_conf.autostart_timeout_seconds
+
+    def disconnect(self):
+        """ Disconnect all services"""
+        for service in self.__all_services:
+            try:
+                service.disconnect()
+            except Exception as e:
+                print("Exception on disconnecting service", e)
