@@ -5,6 +5,7 @@ import json
 import traceback
 import warnings
 from asyncio import sleep
+from threading import Lock
 from typing import Generator, List, Set, Union
 
 # Third Party Libraries
@@ -31,6 +32,11 @@ from af_worklist_manager.models.worklist_revision import WorklistRevision
 from af_worklist_manager.models.worklist_update import WorklistUpdate
 from af_worklist_manager.models.worklist_update_configuration import WorklistUpdateConfiguration
 
+from af_execution_manager import ActivityStartingApi
+from af_worklist_manager import WorklistItem, AfActivityReference
+from af_runtime_service import EbpInstanceReference
+
+from .abstract_service import AbstractService
 from .configuration import Configuration
 from .service_provider import ServiceProvider
 from .worklist_model import Worklist
@@ -41,29 +47,40 @@ def __update_listener(updates: List[ClientWorklistItemUpdate]):
     pass
 
 
+def __worklist_callback_provider(worklist: Worklist) -> str:
+    return None
+
+
 WorklistUpdateListener = type(__update_listener)
 """
     Type for callback function for retrieving worklist updates
 """
 
+WorklistCallbackProvider = type(__worklist_callback_provider)
+"""
+    Type for callback function for providing worklist callback URLs
+"""
 
-class WorklistService(object):
+
+class WorklistService(AbstractService):
     # The fetch size for incremental worklists / updates. If None, the
     # server-side default will be used
     fetch_count: int = None
     __worklist: Worklist = None
     __items: List[ClientWorklistItem] = None
-    __service_provider: ServiceProvider = None
     __worklist_callback: str = None
+    __worklist_callback_provider: WorklistCallbackProvider = None
     __worklist_update_listeners: Set[WorklistUpdateListener] = None
     __af_conf: Configuration = None
     __push_sse_client = None
+    __value_lock: Lock = None
 
     def __init__(self, service_provider: ServiceProvider, configuration: Configuration):
         self.__items = []
-        self.__service_provider = service_provider
         self.__worklist_update_listeners = set()
         self.__af_conf = configuration
+        self.__value_lock = Lock()
+        super().__init__(service_provider=service_provider)
 
     def create_worklist_update_configuration(self, push: bool) -> WorklistUpdateConfiguration:
         """Creates a default worklist update configuration
@@ -83,53 +100,85 @@ class WorklistService(object):
         )
         return wuc
 
-    def get_worklist(self, worklist_callback: str = None) -> List[ClientWorklistItem]:
+    def get_worklist(self, worklist_callback: str = None, worklist_callback_provider: WorklistCallbackProvider = None,
+                     skip_update: bool = False, skip_callback: bool = False) -> List[ClientWorklistItem]:
         """
         Updates and returns the worklist of the current user
         :param str worklist_callback: Optionally an URL implementing the callback endpoint for push updates
+        :param WorklistCallbackProvider worklist_callback_provider: Optionally a callback for generating
+            callback URLs for push updates
+        :param bool skip_update: Skip the update if already connected but no callback options are enabled.
         """
-        if self.__worklist is not None:
-            # simply return the items if push notifications are enabled
-            if self.__worklist_callback is not None:
+        # use a lock to prevent concurrent initialization of the worklist
+        with self.__value_lock:
+            if self.__worklist is not None:
+                # simply return the items if push notifications are enabled
+                if worklist_callback is not None or worklist_callback_provider is not None:
+                    # ensure the callback is still registered
+                    self.enable_callback_updates(worklist_callback=worklist_callback,
+                                                 worklist_callback_provider=worklist_callback_provider)
+                    return self.__items
+                if self.__push_sse_client:
+                    return self.__items
+                # perform update
+                if not skip_update:
+                    return self.update_worklist()
+                else:
+                    return self.__items
+
+            wum: WorklistUpdateManagerApi = self._service_provider.get_service(
+                WorklistUpdateManagerApi
+            )
+            update_conf: WorklistUpdateConfiguration = self.create_worklist_update_configuration(True)
+            wlit: InitialIncClientWorklistData
+            if self.fetch_count is not None:
+                wlit = wum.logon_and_create_client_worklist(body=update_conf, count=self.fetch_count)
+            else:
+                wlit = wum.logon_and_create_client_worklist(body=update_conf)
+
+            # currently no items in the worklist
+            if wlit is None:
                 return self.__items
-            # perform update
-            return self.update_worklist()
 
-        wum: WorklistUpdateManagerApi = self.__service_provider.get_service(
-            WorklistUpdateManagerApi
-        )
-        update_conf: WorklistUpdateConfiguration = self.create_worklist_update_configuration(True)
-        wlit: InitialIncClientWorklistData = None
-        if self.fetch_count is not None:
-            wlit = wum.logon_and_create_client_worklist(body=update_conf, count=self.fetch_count)
-        else:
-            wlit = wum.logon_and_create_client_worklist(body=update_conf)
+            self.__iterate(self.__items, wlit)
+
+            # remember the current worklist meta data
+            self.__worklist = Worklist(
+                wlit.worklist_id, wlit.revision, wlit.client_worklist_id, update_conf, wlit.agent
+            )
+
+            if not skip_callback:
+                self.enable_callback_updates(worklist_callback)
 
-        # currently no items in the worklist
-        if wlit is None:
             return self.__items
 
-        self.__iterate(self.__items, wlit)
-
-        # remember the current worklist meta data
-        self.__worklist = Worklist(
-            wlit.worklist_id, wlit.revision, wlit.client_worklist_id, update_conf, wlit.agent
-        )
-
-        if worklist_callback is not None:
-            callbackData = ClientWorklistRestCallbackData(
-                worklist_callback=worklist_callback,
-                sub_class="ClientWorklistRestCallbackData",
-                id=self.__worklist.worklist_id,
-                client_worklist_id=self.__worklist.client_worklist_id,
-                agent=self.__worklist.agent,
-                revision=self.__worklist.revision,
-                wu_conf=self.__worklist.wu_conf,
+    def enable_callback_updates(self, worklist_callback: str = None,
+                                worklist_callback_provider: WorklistCallbackProvider = None,
+                                client_worklist_id: int = None):
+        """
+        Enable automatic worklist updates using SSE push notifications.
+        """
+        if worklist_callback is not None or worklist_callback_provider is not None:
+            wum: WorklistUpdateManagerApi = self._service_provider.get_service(
+                WorklistUpdateManagerApi
             )
-            wum.register_client_worklist_callback(callbackData)
-            self.__worklist_callback = worklist_callback
-
-        return self.__items
+            is_registered = wum.is_registered_for_updates(self.__worklist.worklist_id, self.__worklist.client_worklist_id)
+            # print(f'Callback for {self.__worklist.worklist_id} registration status: {is_registered}')
+            if not is_registered:
+                if client_worklist_id is None:
+                    client_worklist_id = self.__worklist.client_worklist_id
+                callback_data = ClientWorklistRestCallbackData(
+                    worklist_callback=worklist_callback if worklist_callback else worklist_callback_provider(self.__worklist),
+                    sub_class="ClientWorklistRestCallbackData",
+                    id=self.__worklist.worklist_id,
+                    client_worklist_id=client_worklist_id,
+                    agent=self.__worklist.agent,
+                    revision=self.__worklist.revision,
+                    wu_conf=self.__worklist.wu_conf,
+                )
+                wum.register_client_worklist_callback(callback_data)
+                self.__worklist_callback = worklist_callback
+                self.__worklist_callback_provider = worklist_callback_provider
 
     def enable_push_updates(self):
         """
@@ -145,22 +194,20 @@ class WorklistService(object):
             warnings.warn(
                 "The worklist could not be initialized, probably due to BPM-3581. SSE push will not work."
             )
-        asyncio.run_coroutine_threadsafe(
-            self._process_push_updates(), self.__service_provider.push_event_loop
-        )
+        self._service_provider.thread_pool.submit(self._process_push_updates)
 
-    async def _process_push_updates(self):
+    def _process_push_updates(self):
         """
         Coroutine retrieving SSE push notifications for the worklist, handling registration and reconnects
         """
-        while True:
-            # print("Establishing SSE connection...")
+        while not self._disconnected:
+            print("WorklistService establishing SSE connection...")
             try:
-                sse_connection_id, sse_client = self.__service_provider.connect_sse(
+                sse_connection_id, sse_client = self._service_provider.connect_sse(
                     WorklistUpdateManagerApi
                 )
                 while True:
-                    # print("SSE connection established, registering for worklist push...")
+                    print("WorklistService SSE connection established, registering for worklist push...")
                     callback_data = ClientWorklistSseCallbackData(
                         sse_conn=sse_connection_id,
                         sub_class="ClientWorklistSseCallbackData",
@@ -170,33 +217,27 @@ class WorklistService(object):
                         revision=self.__worklist.revision,
                         wu_conf=self.__worklist.wu_conf,
                     )
-                    wum: WorklistUpdateManagerApi = self.__service_provider.get_service(
+                    wum: WorklistUpdateManagerApi = self._service_provider.get_service(
                         WorklistUpdateManagerApi
                     )
                     wum.register_client_worklist_sse(callback_data)
-                    # print("Worklist registered for SSE push")
+                    print("Worklist registered for SSE push")
                     self.__push_sse_client = sse_client
                     for event in sse_client:
                         if event.event == "SseConnectionEstablished":
                             # print('SSE session was re-established, re-registering..')
                             callback_data.sse_conn = event.data
-                            callback_data.revision = (self.__worklist.revision,)
+                            callback_data.revision = self.__worklist.revision
                             wum.register_client_worklist_sse(callback_data)
                             # print("Worklist registered again for SSE push")
                         elif event.event == "client-worklist-update":
                             # print("Worklist update received")
                             try:
                                 update_dict = json.loads(event.data)
-                                update: WorklistUpdate = self.__service_provider.deserialize(
+                                update: WorklistUpdate = self._service_provider.deserialize(
                                     update_dict, WorklistUpdate
                                 )
-                                self.__apply_worklist_updates(
-                                    update.source_revision,
-                                    update.target_revision,
-                                    update.item_updates,
-                                )
-                                # call the listeners
-                                self._notify_worklist_update_listeners(update.item_updates)
+                                self.apply_worklist_update(update)
                             except Exception as e:
                                 print("Couldn't deserialize and apply update: ", event, e)
                         else:
@@ -204,7 +245,7 @@ class WorklistService(object):
             except ConnectionError:
                 # re-establish connection after some wait time
                 # print("SSE disconnected...")
-                await sleep(self.__af_conf.sse_connect_retry_wait)
+                sleep(self.__af_conf.sse_connect_retry_wait)
             except Exception as e:
                 print("Unknown exception caught during SSE handling", e.__class__)
                 traceback.print_exc()
@@ -264,7 +305,7 @@ class WorklistService(object):
             return
 
         # fetch next
-        inc_cl: IncClientWorklistsApi = self.__service_provider.get_service(IncClientWorklistsApi)
+        inc_cl: IncClientWorklistsApi = self._service_provider.get_service(IncClientWorklistsApi)
         next_it: IncClientWorklistData = inc_cl.inc_client_wl_get_next(inc.inc_wl_id)
         self.__iterate(items, next_it)
 
@@ -273,10 +314,12 @@ class WorklistService(object):
         if self.__worklist is None:
             return self.get_worklist()
 
-        if self.__push_sse_client is not None:
+        if self.__push_sse_client is not None \
+                or self.__worklist_callback is not None\
+                or self.__worklist_callback_provider is not None:
             return self.__items
 
-        wu: WorklistUpdateManagerApi = self.__service_provider.get_service(WorklistUpdateManagerApi)
+        wu: WorklistUpdateManagerApi = self._service_provider.get_service(WorklistUpdateManagerApi)
         inc_updts: InitialIncWorklistUpdateData = wu.get_worklist_updates(
             self.__worklist.worklist_id,
             body=self.__worklist.revision,
@@ -292,6 +335,34 @@ class WorklistService(object):
             self._notify_worklist_update_listeners(updates)
         return self.__items
 
+
+    # TODO define fitting return value
+    def select_activity(
+        self, item: WorklistItem #, signal_handler: SignalHandler = None
+    ) -> bool:
+        """
+        Assigns the current user to the provided activity
+        """
+        asa: ActivityStartingApi = self._service_provider.get_service(
+            ActivityStartingApi
+        )
+        
+        ar: AfActivityReference = item.act_ref
+        ebp_ir = EbpInstanceReference(
+            ar.type,
+            ar.instance_id,
+            ar.instance_log_id,
+            ar.base_template_id,
+            ar.node_id,
+            ar.node_iteration,
+            ar.execution_manager_uris,
+            ar.runtime_manager_uris,
+        )
+
+        asa.select_activity(body=ebp_ir)
+
+        return True
+
     def __iterate_updates(
         self, updates: List[ClientWorklistItemUpdate], inc: IncWorklistUpdateData
     ):
@@ -306,10 +377,21 @@ class WorklistService(object):
             return
 
         # fetch next
-        iwua: IncWorklistUpdateApi = self.__service_provider.get_service(IncWorklistUpdateApi)
+        iwua: IncWorklistUpdateApi = self._service_provider.get_service(IncWorklistUpdateApi)
         next_it: IncWorklistUpdateData = iwua.inc_wl_updt_get_next(inc.inc_upd_id)
         self.__iterate_updates(updates, next_it)
 
+    def apply_worklist_update(self, update: WorklistUpdate):
+        """ Apply the provided worklist update to the worklist
+        """
+        self.__apply_worklist_updates(
+            update.source_revision,
+            update.target_revision,
+            update.item_updates,
+        )
+        # call the listeners
+        self._notify_worklist_update_listeners(update.item_updates)
+
     def __apply_worklist_updates(
         self,
         source_revision: WorklistRevision,
@@ -376,10 +458,38 @@ class WorklistService(object):
     def find_item_by_id(self, item_id: str) -> ClientWorklistItem:
         """Finds a worklist item by its worklist item id. Returns none, if not in the worklist of the user."""
         # print(f'Finding item with id {item_id}')
-        self.update_worklist()
+        self.get_worklist()
         # print(self.__items)
         for item in self.__items:
             if item.id == item_id:
                 # print('Found')
                 return item
         return None
+
+    def find_item_by_ref(self, ebp_ir: EbpInstanceReference) -> ClientWorklistItem:
+        """
+        Finds a worklist item by its activity's EBP Instance Reference.
+        Returns none, if not in the worklist of the user.
+        """
+        self.get_worklist()
+        for item in self.__items:
+            ar: AfActivityReference = item.act_ref
+            if ar.instance_id == ebp_ir.instance_id\
+                    and ar.instance_log_id == ebp_ir.instance_log_id\
+                    and ar.node_id == ebp_ir.node_id\
+                    and ar.node_iteration == ebp_ir.node_iteration:
+                return item
+        return None
+
+
+    def update_worklist_item(self, item: ClientWorklistItem):
+        wum: WorklistUpdateManagerApi = self._service_provider.get_service(
+            WorklistUpdateManagerApi
+        )
+        wum.update_individual_settings(item)
+
+    def disconnect(self):
+        AbstractService.disconnect(self)
+        # close the SSE connection if any
+        if self.__push_sse_client:
+            self.__push_sse_client.close()
